#include <cppshell/shell.hpp>
#include <cppshell/utils/signals.hpp>

#include <sys/stat.h>
#include <unistd.h>

#if defined(__linux__)
#include <wait.h>
#elif defined(__APPLE__)
#include <sys/wait.h>
#else
#error Supported only on Linux and Mac
#endif

#include <cstdio>
#include <stdexcept>

namespace cppshell {
Shell::Shell(const std::string& shell) {
  MakeUniqueDirectory();
  MakeFifoFile();

  int fds_exit_codes[2];
  pipe(fds_exit_codes);

  pid_shell_process_ = fork();
  if (pid_shell_process_ == -1) {
    throw std::logic_error("fork failed");
  }

  if (pid_shell_process_ == 0) {  // child
    close(fds_exit_codes[0]);

    execlp("bash", "bash", command_transmission_file_.data(), NULL);
  } else {  // parent
    fd_exit_codes_transmission_ = fds_exit_codes[1];

    close(fds_exit_codes[1]);
    exit_codes_receiving_ = fdopen(fds_exit_codes[0], "r");

    command_transmission_.open(command_transmission_file_);
  }
}

void Shell::MakeUniqueDirectory() {
  unique_directory_ = "/tmp/cppshell_XXXXXX";
  if (mkdtemp(const_cast<char*>(unique_directory_.data())) == nullptr) {
    throw std::logic_error("make unique directory failed");
  }
}

void Shell::MakeFifoFile() {
  command_transmission_file_ = unique_directory_ + "/commands";
  if (mkfifo(command_transmission_file_.data(), 0600) == -1) {
    throw std::logic_error("make fifo failed");
  }
}

void Shell::ExecuteImpl(const std::string& command, bool need_ignore_sigpipe) {
  if (need_ignore_sigpipe) {
    IgnoreSigpipe();
  }

  command_transmission_ << command << std::endl;

  if (need_ignore_sigpipe) {
    UnignoreSigpipe();
  }
}

int Shell::GetExitCodeLastCommand() {
  Execute("echo $?>&" + std::to_string(fd_exit_codes_transmission_));

  int result;
  int x = fscanf(exit_codes_receiving_, "%d", &result);

  return result;
}

void Shell::Destroy(bool need_ignore_sigpipe) {
  Execute("exit");

  if (need_ignore_sigpipe) {
    IgnoreSigpipe();
  }

  command_transmission_.close();

  if (need_ignore_sigpipe) {
    UnignoreSigpipe();
  }

  waitpid(pid_shell_process_, 0, 0);

  unlink(command_transmission_file_.data());
  rmdir(unique_directory_.data());
}
}  // namespace cppshell
